import gzip
import hashlib
import json
import logging
import random
import string
import threading
import time
import uuid
from typing import Optional, Union
import requests
from requests import Response, Session

from google.protobuf.message import Message
from requests.adapters import HTTPAdapter

from byteplus_rec_core import constant, utils
from byteplus_rec_core.abtract_host_availabler import AbstractHostAvailabler
from byteplus_rec_core.exception import BizException, NetException
from byteplus_rec_core.metrics.metrics import Metrics
from byteplus_rec_core.metrics.metrics_log import MetricsLog
from byteplus_rec_core.option import Option
from byteplus_rec_core.options import Options
from byteplus_rec_core.utils import _milliseconds, HTTPRequest
from byteplus_rec_core.auth import _Credential, _sign

log = logging.getLogger(__name__)

_DEFAULT_PING_URL_FORMAT: str = "{}://{}/predict/api/ping"
_DEFAULT_PING_TIMEOUT_SECONDS: float = 0.5


class Config(object):
    def __init__(self,
                 max_idle_connections: Optional[int] = constant.DEFAULT_MAX_IDLE_CONNECTIONS,
                 keep_alive_ping_interval_seconds: Optional[float] = constant.DEFAULT_KEEPALIVE_PING_INTERVAL_SECONDS):
        self.max_idle_connections = max_idle_connections
        self.keep_alive_ping_interval_seconds = keep_alive_ping_interval_seconds


class _HTTPCaller(object):
    def __init__(self,
                 project_id: str,
                 tenant_id: str,
                 air_auth_token: str,
                 host_availabler: AbstractHostAvailabler,
                 caller_config: Config,
                 schema: str,
                 keep_alive: bool,
                 credential: _Credential = None):
        self._abort: bool = False
        self._project_id = project_id
        self._tenant_id: str = tenant_id
        self._air_auth_token: Optional[str] = air_auth_token
        self._host_availabler = host_availabler
        self._config = caller_config
        self._schema = schema
        self._keep_alive = keep_alive
        self._credential: Optional[_Credential] = credential
        # requests.post creates a new connection for each request, and cannot reuse the connection.
        # Change it to session mode. By default, it maintains a connection pool of up to 10 different hosts.
        self._http_cli: Session = requests.Session()
        self._http_cli.mount("https://", HTTPAdapter(pool_maxsize=self._config.max_idle_connections))
        self._http_cli.mount("http://", HTTPAdapter(pool_maxsize=self._config.max_idle_connections))
        self._local = threading.local()
        if self._keep_alive:
            self._init_heartbeat_executor()

    def _get_req_id(self) -> str:
        return self._local.req_id

    def _init_heartbeat_executor(self):
        threading.Thread(target=self._heartbeat).start()

    def _heartbeat(self):
        if self._abort:
            return
        time.sleep(self._config.keep_alive_ping_interval_seconds)
        for host in self._host_availabler.get_hosts():
            metrics_tags = [
                "from:http_caller",
                "project_id:" + self._project_id,
                "host:" + utils.escape_metrics_tag_value(host)
            ]
            Metrics.counter(constant.METRICS_KEY_HEARTBEAT_COUNT, 1, *metrics_tags)
            utils.ping(self._project_id, self._http_cli, _DEFAULT_PING_URL_FORMAT,
                       self._schema, host, _DEFAULT_PING_TIMEOUT_SECONDS)
        self._heartbeat()

    def do_json_request(self, url: str, request: Union[dict, list], *opts: Option) -> Union[dict, list]:
        options: Options = Option.conv_to_options(opts)
        try:
            req_str: str = json.dumps(request)
        except BaseException as e:
            raise BizException(str(e))
        req_bytes: bytes = req_str.encode("utf-8")
        content_type: str = "application/json"
        rsp_bytes = self._do_request(url, req_bytes, content_type, options)
        try:
            rsp = json.loads(rsp_bytes)
        except BaseException as e:
            metrics_tags = [
                "type:load_json_rsp_fail",
                "project_id:" + self._project_id,
                "url:" + utils.escape_metrics_tag_value(url)
            ]
            Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
            MetricsLog.error(self._get_req_id(), "[ByteplusSDK] load json response fail, project_id:{}, url:{} err:{}",
                             self._project_id, url, e)
            raise BizException(str(e))
        return rsp

    def do_pb_request(self, url: str, request: Message, response: Message, *opts: Option):
        options: Options = Option.conv_to_options(opts)
        req_bytes: bytes = request.SerializeToString()
        content_type: str = "application/x-protobuf"
        rsp_bytes = self._do_request(url, req_bytes, content_type, options)
        try:
            response.ParseFromString(rsp_bytes)
        except BaseException as e:
            metrics_tags = [
                "type:parse_response_fail",
                "project_id:" + self._project_id,
                "url:" + utils.escape_metrics_tag_value(url)
            ]
            Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
            MetricsLog.error(self._get_req_id(), "[ByteplusSDK]parse response fail, project_id:{}, url:{} err:{}",
                             self._project_id, url, e)
            log.error("[ByteplusSDK] parse response fail, url:%s e:%s", url, e)
            raise BizException("parse response fail")

    def _do_request(self, url: str, req_bytes: bytes, content_type: str, options: Options) -> Optional[bytes]:
        req_bytes: bytes = gzip.compress(req_bytes)
        headers: dict = self._build_headers(options, content_type)
        url = self._build_url_with_queries(options, url)
        return self._do_http_request(url, headers, req_bytes, options)

    def _build_headers(self, options: Options, content_type: str) -> dict:
        headers = {
            "Content-Encoding": "gzip",
            # The 'requests' lib support '"Content-Encoding": "gzip"' header,
            # it will decompress gzip response without us
            "Accept-Encoding": "gzip",
            "Content-Type": content_type,
            "Accept": content_type,
            "Tenant-Id": self._tenant_id,
            "Project-Id": self._project_id,
        }
        self._with_options_headers(headers, options)
        return headers

    def _with_options_headers(self, headers: dict, options: Options):
        if options.headers is not None:
            headers.update(options.headers)
        request_id: str = options.request_id
        if request_id is None or len(request_id) == 0:
            request_id = str(uuid.uuid1())
            log.info("[ByteplusSDK] requestID is generated by sdk: '%s'", request_id)
        headers["Request-Id"] = request_id
        self._local.req_id = request_id
        if options.server_timeout is not None:
            headers["Timeout-Millis"] = str(_milliseconds(options.server_timeout))

    @staticmethod
    def _build_url_with_queries(options: Options, url: str):
        queries = {}
        if options.queries is not None:
            queries.update(options.queries)
        if len(queries) == 0:
            return url
        query_parts = []
        for query_name in queries.keys():
            query_parts.append(query_name + "=" + queries[query_name])
        query_string = "&".join(query_parts)
        if "?" in url:
            return url + "&" + query_string
        return url + "?" + query_string

    def _do_http_request(self, url: str, headers: dict, req_bytes: bytes, options: Options) -> Optional[bytes]:
        req = HTTPRequest(headers, url, constant.POST_METHOD_NAME, req_bytes)
        self._with_auth_headers(req)
        start = time.time()
        log.debug("[ByteplusSDK][HTTPCaller] URL:%s, Request Headers:\n%s", url, str(headers))
        try:
            if options.timeout is not None:
                timeout_secs = options.timeout.total_seconds()
                rsp: Response = self._http_cli.post(url=req.url, headers=req.header, data=req.req_bytes,
                                                    timeout=timeout_secs)
            else:
                rsp: Response = self._http_cli.post(url=url, headers=headers, data=req_bytes)
            if rsp.status_code != constant.HTTP_STATUS_OK:
                self._log_err_http_rsp(url, rsp)
                raise BizException("code:{} msg:{}".format(rsp.status_code, rsp.reason))
        except BaseException as e:
            cost = int((time.time() - start) * 1000)
            if self._is_timeout_exception(e):
                metrics_tags = [
                    "type:request_timeout"
                    "project_id:" + self._project_id,
                    "url:" + utils.escape_metrics_tag_value(url),
                ]
                Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
                MetricsLog.error(self._get_req_id(),
                                 "[ByteplusSDK] do http request timeout, project_id:{}, url:{}, cost:{}ms, msg:{}",
                                 self._project_id, url, cost, e)
                log.error("[ByteplusSDK] do http request timeout, url:%s, cost:%dms, msg:%s", url, cost, e)
                raise NetException(str(e))
            metrics_tags = [
                "type:request_occur_exception"
                "project_id:" + self._project_id,
                "url:" + utils.escape_metrics_tag_value(url),
            ]
            Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
            MetricsLog.error(self._get_req_id(),
                             "[ByteplusSDK] do http request occur exception, project_id:{}, url:{}, msg:{}",
                             self._project_id, url, e)
            log.error("[ByteplusSDK] do http request occur io exception, url:%s, cost:%dms, msg:%s", url, cost, e)
            raise BizException(str(e))
        finally:
            cost = int((time.time() - start) * 1000)
            metrics_tags = [
                "project_id:" + self._project_id,
                "url:" + utils.escape_metrics_tag_value(url),
            ]
            Metrics.timer(constant.METRICS_KEY_REQUEST_TOTAL_COST, cost, *metrics_tags)
            Metrics.counter(constant.METRICS_KEY_REQUEST_COUNT, 1, *metrics_tags)
            MetricsLog.info(self._get_req_id(), "[ByteplusSDK] http request, project_id:{}, url:{}, cost:{}ms",
                            self._project_id, url, cost)
            log.debug("[ByteplusSDK] http url:%s, cost:%dms", url, cost)
        return rsp.content

    def _with_auth_headers(self, req: HTTPRequest):
        if self._credential is None:
            self._with_air_auth_headers(req)
            return
        _sign(req, self._credential)

    def _with_air_auth_headers(self, req: HTTPRequest) -> None:
        ts = str(int(time.time()))
        nonce = ''.join(random.sample(string.ascii_letters + string.digits, 8))
        signature = self._cal_signature(req.req_bytes, ts, nonce)
        req.header['Tenant-Ts'] = ts
        req.header['Tenant-Nonce'] = nonce
        req.header['Tenant-Signature'] = signature

    def _cal_signature(self, req_bytes: bytes, ts: str, nonce: str) -> str:
        sha256 = hashlib.sha256()
        sha256.update(self._air_auth_token.encode('utf-8'))
        sha256.update(req_bytes)
        sha256.update(self._tenant_id.encode('utf-8'))
        sha256.update(ts.encode('utf-8'))
        sha256.update(nonce.encode('utf-8'))
        return sha256.hexdigest()

    def _log_err_http_rsp(self, url: str, rsp: Response) -> None:
        metrics_tags = [
            "type:rsp_status_not_ok"
            "project_id:" + self._project_id,
            "url:" + utils.escape_metrics_tag_value(url),
            "status:" + str(rsp.status_code)
        ]
        Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
        rsp_bytes = rsp.content
        if rsp_bytes is not None and len(rsp.content) > 0:
            MetricsLog.error(self._get_req_id(),
                             "[ByteplusSDK] http status not 200, project_id:{}, url:{}, code:{}, msg:{}, headers:\n{}, "
                             "body:\n{}",
                             self._project_id, url, rsp.status_code, rsp.reason, str(rsp.headers), str(rsp_bytes))
            log.error("[ByteplusSDK] http status not 200, url:%s code:%d msg:%s headers:\n%s body:\n%s",
                      url, rsp.status_code, rsp.reason, str(rsp.headers), str(rsp_bytes))
        else:
            MetricsLog.error(self._get_req_id(), "[ByteplusSDK] http status not 200, project_id:{}, url:{}, code:{}, msg:{}, headers:\n{}",
                             self._project_id, url, rsp.status_code, rsp.reason, str(rsp.headers))
            log.error("[ByteplusSDK] http status not 200, url:%s code:%d msg:%s headers:\n%s",
                      url, rsp.status_code, rsp.reason, str(rsp.headers))
        return

    @staticmethod
    def _is_timeout_exception(e):
        lower_err_msg = str(e).lower()
        if "time" in lower_err_msg and "out" in lower_err_msg:
            return True
        return False

    def shutdown(self):
        self._abort = True



class BizException(Exception):
    def __init__(self, msg: str):
        Exception.__init__(self, msg)


class NetException(Exception):
    def __init__(self, msg: str):
        Exception.__init__(self, msg)


from abc import abstractmethod
from typing import List, Optional

from byteplus_rec_core.abtract_host_availabler import AbstractHostAvailabler
from byteplus_rec_core.ping_host_availabler import PingHostAvailabler, Config


# Implement custom HostAvailabler by overriding HostAvailablerFactory.
class HostAvailablerFactory(object):
    @abstractmethod
    def new_host_availabler(self, hosts: List[str],
                            project_id: Optional[str] = "") -> AbstractHostAvailabler:
        return PingHostAvailabler(hosts, project_id, config=None)





import logging
import threading
import time
from threading import Lock
from queue import Queue
from typing import Optional, List

from byteplus_rec_core.metrics.metrics_option import *
from byteplus_rec_core.metrics.constant import *
from byteplus_rec_core.metrics.metrics_reporter import MetricsReporter
from byteplus_rec_core.metrics.protocol import Metric, MetricMessage, MetricLog, MetricLogMessage

log = logging.getLogger(__name__)


class MetricsCollector(object):
    metrics_cfg: MetricsCfg = None
    host_availabler = None
    metrics_reporter: MetricsReporter
    metrics_collector: Queue = None
    metrics_log_collector: Queue = None
    cleaning_metrics_collector: bool = False
    cleaning_metrics_log_collector: bool = False
    initialed: bool = False
    lock: Lock = Lock()

    @classmethod
    def init(cls, cfg: MetricsCfg = None, host_availabler = None):
        if cls.initialed:
            return
        if cfg is None:
            cfg = MetricsCfg()
        cls._do_init(cfg, host_availabler)

    @classmethod
    def init_with_option(cls, *metrics_opts: MetricsOption):
        if cls.initialed:
            return
        cfg = MetricsCfg()
        for opt in metrics_opts:
            opt.fill(cfg)
        cls._do_init(cfg)

    @classmethod
    def _do_init(cls, cfg: MetricsCfg, host_availabler = None):
        cls.lock.acquire()
        if cls.initialed:
            return
        cls.metrics_cfg = cfg
        cls.host_availabler = host_availabler
        # initialize metrics reporter
        cls.metrics_reporter = MetricsReporter(cls.metrics_cfg)
        # initialize metrics collector
        cls.metrics_collector = Queue(maxsize=MAX_METRICS_SIZE)
        cls.metrics_log_collector = Queue(maxsize=MAX_METRICS_LOG_SIZE)

        if not cls.is_enable_metrics() and not cls.is_enable_metrics_log():
            cls.initialed = True
            return
        threading.Thread(target=cls._report).start()
        cls.initialed = True
        cls.lock.release()

    @classmethod
    def is_initialed(cls) -> bool:
        return cls.initialed

    @classmethod
    def is_enable_metrics(cls) -> bool:
        if cls.metrics_cfg is None:
            return False
        return cls.metrics_cfg.enable_metrics

    @classmethod
    def is_enable_metrics_log(cls) -> bool:
        if cls.metrics_cfg is None:
            return False
        return cls.metrics_cfg.enable_metrics_log

    @classmethod
    def emit_metrics(cls, metrics_type: str, name: str, value: int, *tag_kvs: str):
        if not cls.is_enable_metrics():
            return
        # spin when cleaning collector
        try_times = 0
        while cls.cleaning_metrics_collector:
            if try_times >= MAX_SPIN_TIMES:
                return
            # sleep 10ms.
            time.sleep(1/1000 * 10)
            try_times += 1
        if cls.metrics_collector.full():
            log.debug("[BytePlusSDK][Metrics]: The number of metrics exceeds the limit, the metrics write is rejected")
            return
        metric_name: str = name
        if not utils.is_empty_str(cls.metrics_cfg.prefix):
            metric_name = "{}.{}".format(cls.metrics_cfg.prefix, metric_name)
        metric: Metric = Metric()
        metric.name = metric_name
        metric.type = metrics_type
        metric.value = float(value)
        metric.timestamp = utils.current_time_millis()
        metric.tags.update(cls._recover_tags(*tag_kvs))
        cls.metrics_collector.put(metric)

    @classmethod
    def _recover_tags(cls, *tag_kvs: str) -> dict:
        tags = {}
        for tag in tag_kvs:
            kv = tag.split(':', 1)  # only split once, the first part is key, the rest is value
            if len(kv) < 2:
                continue
            tags[kv[0]] = kv[1]
        return tags

    @classmethod
    def emit_log(cls, log_id: str, message: str, log_level: str, timestamp: int):
        if not cls.is_enable_metrics_log():
            return
        # spin when cleaning collector
        try_times = 0
        while cls.cleaning_metrics_log_collector:
            if try_times >= MAX_SPIN_TIMES:
                return
            # sleep 10ms.
            time.sleep(1/1000 * 10)
            try_times += 1
        if cls.metrics_log_collector.full():
            log.debug("[BytePlusSDK][Metrics]: The number of metrics logs exceeds the limit, the metrics write is "
                      "rejected")
            return
        metric_log: MetricLog = MetricLog()
        metric_log.id = log_id
        metric_log.message = message
        metric_log.level = log_level
        metric_log.timestamp = timestamp
        cls.metrics_log_collector.put(metric_log)

    @classmethod
    def _report(cls):
        time.sleep(cls.metrics_cfg.report_interval_seconds)
        if cls.is_enable_metrics():
            cls._report_metrics()
        if cls.is_enable_metrics_log():
            cls._report_metrics_log()
        cls._report()

    @classmethod
    def _report_metrics(cls):
        if cls.metrics_collector.empty():
            return
        metrics: List[Metric] = []
        cls.cleaning_metrics_collector = True
        while True:
            if cls.metrics_collector.empty():
                break
            metrics.append(cls.metrics_collector.get())
        cls.cleaning_metrics_collector = False
        cls._do_report_metrics(metrics)

    @classmethod
    def _do_report_metrics(cls, metrics: List[Metric]):
        url: str = METRICS_URL_FORMAT.format(cls.metrics_cfg.http_schema, cls._get_domain(METRICS_PATH))
        metric_message: MetricMessage = MetricMessage()
        metric_message.metrics.extend(metrics)
        try:
            cls.metrics_reporter.report_metrics(metric_message, url)
        except BaseException as e:
            log.error("[BytePlusSDK][Metrics] report metrics exception, msg:{}, url:{}".format(str(e), url))

    @classmethod
    def _report_metrics_log(cls):
        if cls.metrics_log_collector.empty():
            return
        metric_logs: List[MetricLog] = []
        cls.cleaning_metrics_log_collector = True
        while True:
            if cls.metrics_log_collector.empty():
                break
            metric_logs.append(cls.metrics_log_collector.get())
        cls.cleaning_metrics_log_collector = False
        cls.do_report_metrics_log(metric_logs)

    @classmethod
    def do_report_metrics_log(cls, metric_logs: List[MetricLog]):
        url: str = METRICS_LOG_URL_FORMAT.format(cls.metrics_cfg.http_schema, cls._get_domain(METRICS_LOG_PATH))
        metric_log_message: MetricLogMessage = MetricLogMessage()
        metric_log_message.metric_logs.extend(metric_logs)
        try:
            cls.metrics_reporter.report_metrics_log(metric_log_message, url)
        except BaseException as e:
            log.error("[BytePlusSDK][Metrics] report metrics log exception, msg:{}, url:{}".format(str(e), url))

    @classmethod
    def _get_domain(cls, path: str) -> str:
        if cls.host_availabler is None:
            return cls.metrics_cfg.domain
        return cls.host_availabler.get_host(path)



import time
import logging
import threading
import uuid

from byteplus_rec_core.metrics.metrics import Metrics
from byteplus_rec_core.metrics.metrics_collector import MetricsCollector
from byteplus_rec_core.metrics.metrics_log import MetricsLog
from byteplus_rec_core.metrics.metrics_option import MetricsOption


class MetricsExample(object):
    times = 100

    @classmethod
    def metrics_init(cls):
        # default log level is warning
        logging.basicConfig(level=logging.DEBUG)
        MetricsCollector.init_with_option(
            MetricsOption.with_metrics_http_schema("http"),
            MetricsOption.enable_metrics(),
            MetricsOption.enable_metrics_log(),
            MetricsOption.with_report_interval_seconds(5),
            MetricsOption.with_metrics_prefix("test.byteplus.sdk"),
            MetricsOption.with_metrics_timeout_seconds(1),
            MetricsOption.with_metrics_domain("rec-b-ap-singapore-1.byteplusapi.com")
        )

    # test demo for store report
    @classmethod
    def store_report(cls):
        print("start store reporting...")
        for i in range(cls.times):
            Metrics.store(
                "java.request.store", 200, "type:test_metrics1", "url:https://asfwe.sds.com/test?qu1=xxx&qu2=yyy")
            Metrics.store(
                "java.request.store", 100, "type:test_metrics2", "url:https://asfwe.sds.com/test?qu1=xxx&qu2=yyy")
            Metrics.store(
                "java.request.store", 200, "type:test_metrics3", "url:https://asfwe.sds.com/test?qu1=xxx&qu2=yyy")
            Metrics.store(
                "java.request.store", 100, "type:test_metrics4", "url:https://asfwe.sds.com/test?qu1<eq>xxx&qu2<eq>yyy")
            # 20s
            time.sleep(20)
        print("stop store reporting")

    # test demo for counter report
    @classmethod
    def counter_report(cls):
        print("start counter reporting...")
        for i in range(cls.times):
            Metrics.counter("java.request.counter", 1, "type:test_counter1")
            Metrics.counter("java.request.counter", 1, "type:test_counter2")
            # 20s
            time.sleep(20)
        print("stop counter reporting")

    # test demo for timer report
    @classmethod
    def timer_report(cls):
        print("start timer reporting...")
        for i in range(cls.times):
            Metrics.timer("java.request.timer", 140, "type:test_timer4")
            Metrics.timer("java.request.timer", 160, "type:test_timer4")
            # 30s
            time.sleep(30)
        print("stop timer reporting")


if __name__ == '__main__':
    MetricsExample.metrics_init()
    # MetricsExample.store_report()
    # MetricsExample.counter_report()
    MetricsExample.timer_report()
    # log_id = str(uuid.uuid4())
    # print("log_id: {}".format(log_id))
    # MetricsLog.info(log_id, "this is a test log: name:{}, value: {}", "demo", 1)




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sdk_metrics.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='sdk_metrics.proto',
  package='metric',
  syntax='proto3',
  serialized_options=b'\n&com.byteplus.rec.core.metrics.protocolZAgithub.com/byteplus-sdk/byteplus-sdk-go-rec-core/metrics/protocol',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11sdk_metrics.proto\x12\x06metric\"\x9b\x01\n\x06Metric\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\x12\r\n\x05value\x18\x04 \x01(\x01\x12&\n\x04tags\x18\x05 \x03(\x0b\x32\x18.metric.Metric.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"0\n\rMetricMessage\x12\x1f\n\x07metrics\x18\x01 \x03(\x0b\x32\x0e.metric.Metric\"J\n\tMetricLog\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05level\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\x12\x0f\n\x07message\x18\x04 \x01(\t\":\n\x10MetricLogMessage\x12&\n\x0bmetric_logs\x18\x01 \x03(\x0b\x32\x11.metric.MetricLogBk\n&com.byteplus.rec.core.metrics.protocolZAgithub.com/byteplus-sdk/byteplus-sdk-go-rec-core/metrics/protocolb\x06proto3'
)




_METRIC_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='metric.Metric.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='metric.Metric.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='metric.Metric.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=142,
  serialized_end=185,
)

_METRIC = _descriptor.Descriptor(
  name='Metric',
  full_name='metric.Metric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='metric.Metric.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='metric.Metric.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='metric.Metric.timestamp', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='metric.Metric.value', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='metric.Metric.tags', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_METRIC_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=185,
)


_METRICMESSAGE = _descriptor.Descriptor(
  name='MetricMessage',
  full_name='metric.MetricMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metrics', full_name='metric.MetricMessage.metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=187,
  serialized_end=235,
)


_METRICLOG = _descriptor.Descriptor(
  name='MetricLog',
  full_name='metric.MetricLog',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='metric.MetricLog.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='level', full_name='metric.MetricLog.level', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='metric.MetricLog.timestamp', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='metric.MetricLog.message', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=237,
  serialized_end=311,
)


_METRICLOGMESSAGE = _descriptor.Descriptor(
  name='MetricLogMessage',
  full_name='metric.MetricLogMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric_logs', full_name='metric.MetricLogMessage.metric_logs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=313,
  serialized_end=371,
)

_METRIC_TAGSENTRY.containing_type = _METRIC
_METRIC.fields_by_name['tags'].message_type = _METRIC_TAGSENTRY
_METRICMESSAGE.fields_by_name['metrics'].message_type = _METRIC
_METRICLOGMESSAGE.fields_by_name['metric_logs'].message_type = _METRICLOG
DESCRIPTOR.message_types_by_name['Metric'] = _METRIC
DESCRIPTOR.message_types_by_name['MetricMessage'] = _METRICMESSAGE
DESCRIPTOR.message_types_by_name['MetricLog'] = _METRICLOG
DESCRIPTOR.message_types_by_name['MetricLogMessage'] = _METRICLOGMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _METRIC_TAGSENTRY,
    '__module__' : 'sdk_metrics_pb2'
    # @@protoc_insertion_point(class_scope:metric.Metric.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'sdk_metrics_pb2'
  # @@protoc_insertion_point(class_scope:metric.Metric)
  })
_sym_db.RegisterMessage(Metric)
_sym_db.RegisterMessage(Metric.TagsEntry)

MetricMessage = _reflection.GeneratedProtocolMessageType('MetricMessage', (_message.Message,), {
  'DESCRIPTOR' : _METRICMESSAGE,
  '__module__' : 'sdk_metrics_pb2'
  # @@protoc_insertion_point(class_scope:metric.MetricMessage)
  })
_sym_db.RegisterMessage(MetricMessage)

MetricLog = _reflection.GeneratedProtocolMessageType('MetricLog', (_message.Message,), {
  'DESCRIPTOR' : _METRICLOG,
  '__module__' : 'sdk_metrics_pb2'
  # @@protoc_insertion_point(class_scope:metric.MetricLog)
  })
_sym_db.RegisterMessage(MetricLog)

MetricLogMessage = _reflection.GeneratedProtocolMessageType('MetricLogMessage', (_message.Message,), {
  'DESCRIPTOR' : _METRICLOGMESSAGE,
  '__module__' : 'sdk_metrics_pb2'
  # @@protoc_insertion_point(class_scope:metric.MetricLogMessage)
  })
_sym_db.RegisterMessage(MetricLogMessage)


DESCRIPTOR._options = None
_METRIC_TAGSENTRY._options = None
# @@protoc_insertion_point(module_scope)



import datetime
import hashlib
import hmac
from typing import Dict, Optional, List, Tuple
from urllib.parse import urlparse, parse_qsl, urlencode, quote

from byteplus_rec_core.utils import HTTPRequest

_TIME_FORMAT_V4 = "%Y%m%dT%H%M%SZ"


class _Credential(object):
    def __init__(self, ak: str, sk: str, service: str, region: str, session_token: Optional[str] = None):
        self.access_key_id = ak
        self.secret_access_key = sk
        self.service = service
        self.region = region
        self.session_token = session_token


class _Metadata(object):
    def __init__(self,
                 service: str,
                 region: str,
                 algorithm: Optional[str] = None,
                 credential_scope: Optional[str] = None,
                 signed_headers: Optional[str] = None,
                 date: Optional[str] = None):
        self.service = service
        self.region = region
        self.algorithm = algorithm
        self.credential_scope = credential_scope
        self.signed_headers = signed_headers
        self.date = date


def _now() -> datetime.datetime:
    return datetime.datetime.utcnow()


def _timestamp_v4() -> str:
    return _now().strftime(_TIME_FORMAT_V4)


def _prepare_request_v4(req: HTTPRequest):
    necessary_defaults: Dict[str, str] = {
        "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
        "X-Date": _timestamp_v4()
    }
    for header, value in necessary_defaults.items():
        if len(req.header.get(header, "")) == 0:
            req.header[header] = value

    path = urlparse(req.url).path
    if len(path) == 0:
        req.url = req.url + "/"


def _sign(req: HTTPRequest, cred: _Credential):
    _prepare_request_v4(req)
    meta: _Metadata = _Metadata(cred.service, cred.region)

    # Task1
    hashed_canon_req: str = _hashed_canonical_request_v4(req, meta)

    # Task2
    string_to_sign_ret: str = _string_to_sign(req.header, hashed_canon_req, meta)

    # Task3
    signing_key_ret = _signing_key(cred.secret_access_key, meta.date, meta.region, meta.service)
    signature_ret = _signature(signing_key_ret, string_to_sign_ret)

    req.header["Authorization"] = _build_auth_header(signature_ret, meta, cred)

    if cred.session_token:
        req.header['X-Security-Token'] = cred.session_token


def _build_auth_header(signature_ret: str, meta: _Metadata, cred: _Credential) -> str:
    credential: str = cred.access_key_id + "/" + meta.credential_scope
    return meta.algorithm + \
           " Credential=" + credential + \
           ", SignedHeaders=" + meta.signed_headers + \
           ", Signature=" + \
           signature_ret


def _hashed_canonical_request_v4(req: HTTPRequest, meta: _Metadata) -> str:
    parse_result = urlparse(req.url)
    # encode body and generate body hash
    if req.req_bytes is not None:
        content_hash = hashlib.sha256(req.req_bytes)
    else:
        content_hash = hashlib.sha256(b'')
    req.header["X-Content-Sha256"] = content_hash.hexdigest()
    req.header["Host"] = parse_result.netloc

    sorted_headers: List[Tuple[str, str]] = []
    allowed_headers = ("content-type", "content-md5", "host")
    for key, value in req.header.items():
        key_lower = key.lower()
        if key_lower in allowed_headers or key_lower.startswith("x-"):
            sorted_headers.append((key_lower, value.strip()))
    sorted_headers.sort(key=lambda x: x[0])

    headers_to_sign: str = ""
    sorted_header_keys: List[str] = []
    for key_lower, value in sorted_headers:
        if key_lower == "host":
            split = value.split(":")
            # host is ip:port, make value = ip
            if len(split) == 2 and (split[1] == "80" or split[1] == "443"):
                value = split[0]
        headers_to_sign += key_lower + ":" + value + "\n"
        sorted_header_keys.append(key_lower)
    meta.signed_headers = ";".join(sorted_header_keys)

    # keep k,v order with server
    sorted_queries: List = parse_qsl(parse_result.query)
    sorted_queries.sort(key=lambda x: x[0])
    query_str: str = urlencode(sorted_queries)

    safe_chars = '/~'
    req_parts = [req.method.upper(), quote(parse_result.path, safe_chars), _normal_query(query_str), headers_to_sign,
                 meta.signed_headers, content_hash.hexdigest()]
    canonical_request = '\n'.join(req_parts)

    return hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()


def _normal_query(query_str: str) -> str:
    return query_str.replace("+", "%20", -1)


def _string_to_sign(header: dict, hashed_canon_req: str, meta: _Metadata) -> str:
    request_ts = header.get("X-Date")
    meta.algorithm = "HMAC-SHA256"
    meta.date = _ts_date(request_ts)
    meta.credential_scope = "/".join([meta.date, meta.region, meta.service, "request"])
    return "\n".join([meta.algorithm, request_ts, meta.credential_scope, hashed_canon_req])


def _ts_date(timestamp: str) -> str:
    return timestamp[:8]


def _signing_key(secret_access_key: str, date: str, region: str, service: str) -> bytes:
    k_date = hmac.new(secret_access_key.encode('utf-8'), date.encode('utf-8'), hashlib.sha256).digest()
    k_region = hmac.new(k_date, region.encode('utf-8'), hashlib.sha256).digest()
    k_service = hmac.new(k_region, service.encode('utf-8'), hashlib.sha256).digest()
    k_signing = hmac.new(k_service, "request".encode('utf-8'), hashlib.sha256).digest()
    return k_signing


def _signature(signing_key_ret: bytes, string_to_sign_ret: str) -> str:
    hsh = hmac.new(signing_key_ret, string_to_sign_ret.encode('utf-8'), hashlib.sha256)
    return hsh.hexdigest()



import time
import uuid
from datetime import timedelta
import logging
from typing import List, Optional

from byteplus_rec_core import constant
from requests import Session, Response

from byteplus_rec_core.exception import NetException, BizException
from byteplus_rec_core.metrics.metrics_log import MetricsLog

log = logging.getLogger(__name__)


def current_time_millis() -> int:
    return int(time.time() * 1000)


def _milliseconds(delta: timedelta) -> int:
    return int(delta.total_seconds() * 1000.0)


def do_with_retry(call, request, opts: tuple, retry_times: int):
    # To ensure the request is successfully received by the server,
    # it should be retried after a network exception occurs.
    # To prevent the retry from causing duplicate uploading same data,
    # the request should be retried by using the same requestId.
    # If a new requestId is used, it will be treated as a new request
    # by the server, which may save duplicate data
    if retry_times < 0:
        retry_times = 0
    try_times = retry_times + 1
    for i in range(try_times):
        try:
            rsp = call(request, *opts)
        except NetException as e:
            if i == try_times - 1:
                log.error("[DoRetryRequest] fail finally after retried '%s' times", try_times)
                raise BizException(str(e))
            continue
        return rsp
    return


def build_url(schema: str, host: str, path: str) -> str:
    if path[0] == '/':
        return "{}://{}{}".format(schema, host, path)
    return "{}://{}/{}".format(schema, host, path)


def none_empty_str(st: List[str]) -> bool:
    if str is None:
        return False
    for s in st:
        if s is None or len(s) == 0:
            return False
    return True


def is_all_empty_str(st: List[str]) -> bool:
    if st is None:
        return True
    for s in st:
        if s is not None and len(s) > 0:
            return False
    return True


def is_empty_str(st: str) -> bool:
    return st is None or len(st) == 0


def ping(project_id: str, http_cli: Session, ping_url_format: str,
         schema: str, host: str, ping_timeout_seconds: float) -> bool:
    url: str = ping_url_format.format(schema, host)
    req_id: str = "ping_" + str(uuid.uuid1())
    headers = {
        "Request-Id": req_id,
        "Project-Id": project_id,
    }
    start = time.time()
    try:
        rsp: Response = http_cli.get(url, headers=headers, timeout=ping_timeout_seconds)
        cost = int((time.time() - start) * 1000)
        if is_ping_success(rsp):
            MetricsLog.info(req_id, "[ByteplusSDK] ping success, project_id:{}, host:{}, cost:{}ms",
                            project_id, host, cost)
            log.debug("[ByteplusSDK] ping success, host:'%s' cost:%dms", host, cost)
            return True
        MetricsLog.warn(req_id, "[ByteplusSDK] ping fail, project_id:{}, host:{}, cost:{}ms, status:{}",
                        project_id, host, cost, rsp.status_code)
        log.warning("[ByteplusSDK] ping fail, host:'%s', cost:%dms, status:'%s'", host, cost, rsp.status_code)
        return False
    except BaseException as e:
        cost = int((time.time() - start) * 1000)
        MetricsLog.warn(req_id, "[ByteplusSDK] ping find err, project_id:{}, host:{}, cost:{}ms, err:{}",
                        project_id, host, cost, e)
        log.warning("[ByteplusSDK] ping find err, host:'%s', cost:%dms, err:'%s'", host, cost, e)
        return False


def is_ping_success(rsp: Response) -> bool:
    if rsp.status_code != constant.HTTP_STATUS_OK:
        return False
    if rsp.content is None:
        return False
    rsp_str: str = str(rsp.content)
    return len(rsp_str) < 20 and "pong" in rsp_str


def is_timeout_exception(e):
    lower_err_msg = str(e).lower()
    if "time" in lower_err_msg and "out" in lower_err_msg:
        return True
    return False


def escape_metrics_tag_value(value: str) -> str:
    value = value.replace("?", "-qu-", -1)
    value = value.replace("&", "-and-", -1)
    value = value.replace("=", "-eq-", -1)
    return value


class HTTPRequest(object):
    def __init__(self,
                 header: dict,
                 url: str,
                 method: str,
                 req_bytes: bytes):
        self.header: dict = header
        self.url: str = url
        self.method: str = method
        self.req_bytes: bytes = req_bytes



__title__ = 'byteplus-rec-core'
__description__ = 'core pkg for byteplus rec'
__url__ = 'https://github.com/byteplus-sdk'
__version__ = '1.0.3'
__author__ = 'Guowei Lei'
__author_email__ = 'leiguowei@bytedance.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2021 Bytedance'



import threading
from typing import Optional, Union, List
import logging

from google.protobuf.message import Message

from byteplus_rec_core.abtract_host_availabler import AbstractHostAvailabler
from byteplus_rec_core.host_availabler_factory import HostAvailablerFactory
from byteplus_rec_core.http_caller import Config as HTTPCallerConfig
from byteplus_rec_core.http_caller import _HTTPCaller
from byteplus_rec_core.metrics.metrics_collector import MetricsCollector
from byteplus_rec_core.metrics.metrics_option import MetricsCfg
from byteplus_rec_core.option import Option
from byteplus_rec_core.abstract_region import AbstractRegion
from byteplus_rec_core.auth import _Credential
from byteplus_rec_core import utils

log = logging.getLogger(__name__)


class HTTPClient(object):
    def __init__(self, http_caller: _HTTPCaller, host_availabler: AbstractHostAvailabler, schema: str, project_id: str):
        self._http_caller = http_caller
        self._host_availabler = host_availabler
        self._schema = schema
        self._project_id = project_id

    def do_pb_request(self, path: str, request: Message, response: Message, *opts: Option):
        self._http_caller.do_pb_request(self._build_url(path), request, response, *opts)

    def do_json_request(self, path: str, request: Union[dict, list], *opts: Option) -> Union[dict, list]:
        return self._http_caller.do_json_request(self._build_url(path), request, *opts)

    def _build_url(self, path: str):
        host: str = self._host_availabler.get_host(path)
        return utils.build_url(self._schema, host, path)

    def shutdown(self):
        self._host_availabler.shutdown()
        self._http_caller.shutdown()


_global_host_availabler_lock = threading.Lock()
_global_host_availabler: Optional[AbstractHostAvailabler] = None


class _HTTPClientBuilder(object):
    def __init__(self):
        self._tenant_id: Optional[str] = None
        self._project_id: Optional[str] = None
        self._use_air_auth: Optional[bool] = False
        self._air_auth_token: Optional[str] = None
        self._auth_ak: Optional[str] = None
        self._auth_sk: Optional[str] = None
        self._auth_service: Optional[str] = None
        self._schema: Optional[str] = None
        self._hosts: Optional[List[str]] = None
        self._region: Optional[AbstractRegion] = None
        self._host_availabler_factory: Optional[HostAvailablerFactory] = None
        self._keep_alive: Optional[bool] = False
        self._caller_config: Optional[HTTPCallerConfig] = None
        self._host_availabler: Optional[AbstractHostAvailabler] = None
        self._metrics_cfg: Optional[MetricsCfg] = None

    def tenant_id(self, tenant_id: str):
        self._tenant_id = tenant_id
        return self

    def project_id(self, project_id: str):
        self._project_id = project_id
        return self

    def use_air_auth(self, use_air_auth: bool):
        self._use_air_auth = use_air_auth
        return self

    def air_auth_token(self, air_auth_token: str):
        self._air_auth_token = air_auth_token
        return self

    def auth_ak(self, ak: str):
        self._auth_ak = ak
        return self

    def auth_sk(self, sk: str):
        self._auth_sk = sk
        return self

    def auth_service(self, auth_service: str):
        self._auth_service = auth_service
        return self

    def schema(self, schema: str):
        self._schema = schema
        return self

    def hosts(self, hosts: list):
        self._hosts = hosts
        return self

    def region(self, region: AbstractRegion):
        self._region = region
        return self

    def host_availabler_factory(self, host_availabler_factory: HostAvailablerFactory):
        self._host_availabler_factory = host_availabler_factory
        return self

    def keep_alive(self, keep_alive: bool):
        self._keep_alive = keep_alive
        return self

    def caller_config(self, caller_config: HTTPCallerConfig):
        self._caller_config = caller_config
        return self

    def metrics_cfg(self, metrics_cfg: MetricsCfg):
        self._metrics_cfg = metrics_cfg
        return self

    def build(self) -> HTTPClient:
        global _global_host_availabler

        self._check_required_field()
        self._fill_default()
        if not MetricsCollector.is_initialed() and self._metrics_cfg is not None:
            if self._metrics_cfg.enable_metrics or self._metrics_cfg.enable_metrics_log:
                self._init_global_host_availabler()
        MetricsCollector.init(self._metrics_cfg, _global_host_availabler)
        return HTTPClient(self._new_http_caller(), self._host_availabler, self._schema, self._project_id)

    def _check_required_field(self):
        if self._tenant_id is None or len(self._tenant_id) == 0:
            raise Exception("tenant id is null")
        self._check_auth_required_field()
        if self._region is None:
            raise Exception("region is null")

    def _check_auth_required_field(self):
        if self._use_air_auth:
            if utils.is_empty_str(self._air_auth_token):
                raise Exception("token cannot be null")
            return

        if utils.is_all_empty_str([self._auth_ak, self._auth_sk]):
            raise Exception("ak and sk cannot be null")

    def _fill_default(self):
        if utils.is_empty_str(self._schema):
            self._schema = "https"
        # # fill hostAvailabler.
        if self._host_availabler_factory is None:
            self._host_availabler_factory = HostAvailablerFactory()
        self._host_availabler: AbstractHostAvailabler = self._new_host_availabler()

        # fill default caller config.
        if self._caller_config is None:
            self._caller_config = HTTPCallerConfig()

    def _new_host_availabler(self) -> AbstractHostAvailabler:
        if self._hosts is not None and len(self._hosts) > 0:
            return self._host_availabler_factory.new_host_availabler(hosts=self._hosts)
        return self._host_availabler_factory.new_host_availabler(hosts=self._region.get_hosts(),
                                                                 project_id=self._project_id)

    def _init_global_host_availabler(self):
        global _global_host_availabler
        global _global_host_availabler_lock

        _global_host_availabler_lock.acquire()
        if _global_host_availabler is not None:
            _global_host_availabler_lock.release()
            return
        _global_host_availabler = self._new_host_availabler()
        _global_host_availabler_lock.release()

    def _new_http_caller(self) -> _HTTPCaller:
        if self._use_air_auth:
            return _HTTPCaller(
                self._project_id,
                self._tenant_id,
                self._air_auth_token,
                self._host_availabler,
                self._caller_config,
                self._schema,
                self._keep_alive
            )
        credential: _Credential = _Credential(
            self._auth_ak,
            self._auth_sk,
            self._auth_service,
            self._region.get_auth_region(),
        )
        _http_caller: _HTTPCaller = _HTTPCaller(
            self._project_id,
            self._tenant_id,
            self._air_auth_token,
            self._host_availabler,
            self._caller_config,
            self._schema,
            self._keep_alive,
            credential
        )
        return _http_caller


def new_http_client_builder() -> _HTTPClientBuilder:
    return _HTTPClientBuilder()



from byteplus_rec_core import utils

from byteplus_rec_core.metrics import constant
from byteplus_rec_core.metrics.metrics_collector import MetricsCollector


class Metrics(object):
    # description: Store tagKvs should be formatted as "key:value"
    # example: store("goroutine.count", 400, "ip:127.0.0.1")
    @staticmethod
    def store(key: str, value: int, *tag_kvs: str):
        MetricsCollector.emit_metrics(constant.METRICS_TYPE_STORE, key, value, *tag_kvs)

    # description: Store tagKvs should be formatted as "key:value"
    # example: counter("request.count", 1, "method:user", "type:upload")
    @staticmethod
    def counter(key: str, value: int, *tag_kvs: str):
        MetricsCollector.emit_metrics(constant.METRICS_TYPE_COUNTER, key, value, *tag_kvs)

    # The unit of `value` is milliseconds
    # example: timer("request.cost", 100, "method:user", "type:upload")
    # description: Store tagKvs should be formatted as "key:value"
    @staticmethod
    def timer(key: str, value: int, *tag_kvs: str):
        MetricsCollector.emit_metrics(constant.METRICS_TYPE_TIMER, key, value, *tag_kvs)

    # The unit of `begin` is milliseconds
    # example: latency("request.latency", startTime, "method:user", "type:upload")
    # description: Store tagKvs should be formatted as "key:value"
    @staticmethod
    def latency(key: str, begin: int, *tag_kvs: str):
        MetricsCollector.emit_metrics(constant.METRICS_TYPE_TIMER, key,
                                      utils.current_time_millis() - begin, *tag_kvs)

    # description: Store tagKvs should be formatted as "key:value"
    # example: rateCounter("request.count", 1, "method:user", "type:upload")
    @staticmethod
    def rate_counter(key: str, value: int, *tag_kvs: str):
        MetricsCollector.emit_metrics(constant.METRICS_TYPE_RATE_COUNTER, key, value, *tag_kvs)

    # description:
    #   - meter(xx) = counter(xx) + rateCounter(xx.rate)
    #   - Store tagKvs should be formatted as "key:value"
    # example: rateCounter("request.count", 1, "method:user", "type:upload")
    @staticmethod
    def meter(key: str, value: int, *tag_kvs: str):
        MetricsCollector.emit_metrics(constant.METRICS_TYPE_METER, key, value, *tag_kvs)



from enum import Enum, unique

# metrics default domain and prefix
DEFAULT_METRICS_DOMAIN: str = "rec-api-sg1.recplusapi.com"
DEFAULT_METRICS_PREFIX: str = "byteplus.rec.sdk"
DEFAULT_HTTP_SCHEMA: str = "https"

# monitor url format
METRICS_URL_FORMAT: str = "{}://{}/predict/api/monitor/metrics"
METRICS_LOG_URL_FORMAT: str = "{}://{}/predict/api/monitor/metrics/log"

# domain path
METRICS_PATH: str = "/monitor/metrics"
METRICS_LOG_PATH: str = "/monitor/metrics/log"

# metrics flush interval
DEFAULT_REPORT_INTERVAL_SECONDS: float = 15
DEFAULT_HTTP_TIMEOUT_SECONDS: float = 0.8
MAX_TRY_TIMES: int = 3
MAX_SPIN_TIMES: int = 5
SUCCESS_HTTP_CODE: int = 200
MAX_METRICS_SIZE: int = 10000
MAX_METRICS_LOG_SIZE: int = 5000

# metrics log level
LOG_LEVEL_TRACE: str = "trace"
LOG_LEVEL_DEBUG: str = "debug"
LOG_LEVEL_INFO: str = "info"
LOG_LEVEL_NOTICE: str = "notice"
LOG_LEVEL_WARN: str = "warn"
LOG_LEVEL_ERROR: str = "error"
LOG_LEVEL_FATAL: str = "fatal"

# metrics type
METRICS_TYPE_COUNTER: str = "counter"
METRICS_TYPE_STORE: str = "store"
METRICS_TYPE_TIMER: str = "timer"
METRICS_TYPE_RATE_COUNTER: str = "rate_counter"
METRICS_TYPE_METER: str = "meter"



from byteplus_rec_core.constant import STATUS_CODE_SUCCESS, STATUS_CODE_TOO_MANY_REQUEST, STATUS_CODE_OPERATION_LOSS, \
    STATUS_CODE_IDEMPOTENT


def is_upload_success(code: int) -> bool:
    return code == STATUS_CODE_SUCCESS or code == STATUS_CODE_IDEMPOTENT


def is_success(code: int) -> bool:
    return code == STATUS_CODE_SUCCESS or code == 200


def is_server_overload(code: int) -> bool:
    return code == STATUS_CODE_TOO_MANY_REQUEST


def is_loss_operation(code: int) -> bool:
    return code == STATUS_CODE_OPERATION_LOSS



HTTP_STATUS_OK: int = 200

HTTP_STATUS_NOT_FOUND: int = 404

# The request was executed successfully without any exception
STATUS_CODE_SUCCESS: int = 0
# A Request with the same "Request-ID" was already received. This Request was rejected
STATUS_CODE_IDEMPOTENT: int = 409
# Operation information is missing due to an unknown exception
STATUS_CODE_OPERATION_LOSS: int = 410
# The server hope slow down request frequency, and this request was rejected
STATUS_CODE_TOO_MANY_REQUEST: int = 429

POST_METHOD_NAME = "POST"


# The default max idle connections of okhttp client connection pool
DEFAULT_MAX_IDLE_CONNECTIONS: int = 32
# The default keepalive ping interval
DEFAULT_KEEPALIVE_PING_INTERVAL_SECONDS: float = 45

# Metrics Key
METRICS_KEY_COMMON_INFO = "common.info"
METRICS_KEY_COMMON_WARN = "common.warn"
METRICS_KEY_COMMON_ERROR = "common.err"
METRICS_KEY_REQUEST_TOTAL_COST = "request.total.cost"
METRICS_KEY_REQUEST_COUNT = "request.count"
METRICS_KEY_HEARTBEAT_COUNT = "heartbeat.count"



import requests
from byteplus_rec_core import utils
from requests import Response

from byteplus_rec_core.exception import BizException
from byteplus_rec_core.metrics import constant
from byteplus_rec_core.metrics.metrics_collector import MetricsCfg
from byteplus_rec_core.metrics.protocol import MetricLogMessage, MetricMessage


class MetricsReporter(object):
    def __init__(self, metrics_cfg: MetricsCfg):
        self._http_cli = requests.Session()
        self._metrics_cfg = metrics_cfg

    def report_metrics(self, metric_message: MetricMessage, url: str):
        req_bytes: bytes = metric_message.SerializeToString()
        headers: dict = self._build_metrics_headers()
        self._do_request(url, req_bytes, headers)

    def report_metrics_log(self, metric_log_message: MetricLogMessage, url: str):
        req_bytes: bytes = metric_log_message.SerializeToString()
        headers: dict = self._build_metrics_headers()
        self._do_request(url, req_bytes, headers)

    def _do_request(self, url: str, req_bytes: bytes, headers: dict):
        for i in range(constant.MAX_TRY_TIMES):
            try:
                response: Response = self._http_cli.post(url=url, headers=headers, data=req_bytes,
                                                         timeout=self._metrics_cfg.http_timeout_seconds)
                if response.status_code == constant.SUCCESS_HTTP_CODE:
                    return
                if response.content is None:
                    raise BizException("rsp body is null")
                raise BizException("do http request fail, url:{}, code:{}, rsp:{}".
                                   format(url, response.status_code, response.content))
            except BaseException as e:
                if utils.is_timeout_exception(e) and i < constant.MAX_TRY_TIMES - 1:
                    continue
                raise BizException(str(e))

    @staticmethod
    def _build_metrics_headers() -> dict:
        headers = {
            "Content-Type": "application/x-protobuf",
            "Accept": "application/json"
        }
        return headers



from abc import abstractmethod

from byteplus_rec_core import utils
from byteplus_rec_core.metrics.constant import *


class MetricsCfg(object):
    def __init__(self, enable_metrics: bool = False, enable_metrics_log: bool = False,
                 domain: str = DEFAULT_METRICS_DOMAIN, prefix: str = DEFAULT_METRICS_PREFIX,
                 http_schema: str = DEFAULT_HTTP_SCHEMA,
                 report_interval_seconds: float = DEFAULT_REPORT_INTERVAL_SECONDS,
                 http_timeout_seconds: float = DEFAULT_HTTP_TIMEOUT_SECONDS):
        # When metrics are enabled, monitoring metrics will be reported to the byteplus server during use.
        self.enable_metrics: bool = enable_metrics
        # When metrics log is enabled, the log will be reported to the byteplus server during use.
        self.enable_metrics_log: bool = enable_metrics_log
        # The address of the byteplus metrics service, will be consistent with the host maintained by hostAvailabler.
        self.domain: str = domain
        # The prefix of the Metrics indicator, the default is byteplus.rec.sdk, do not modify.
        self.prefix: str = prefix
        # Use this httpSchema to report metrics to byteplus server, default is https.
        self.http_schema: str = http_schema
        # The reporting interval, the default is 15s, if the QPS is high,
        # the reporting interval can be reduced to prevent data loss.
        self.report_interval_seconds: float = report_interval_seconds
        # Timeout for request reporting.
        self.http_timeout_seconds: float = http_timeout_seconds


class MetricsOption(object):
    @abstractmethod
    def fill(self, cfg: MetricsCfg) -> None:
        raise NotImplementedError

    @staticmethod
    def with_metrics_domain(domain: str):
        class OptionImpl(MetricsOption):
            def fill(self, cfg: MetricsCfg) -> None:
                if not utils.is_empty_str(domain):
                    cfg.domain = domain

        return OptionImpl()

    @staticmethod
    def with_metrics_prefix(prefix: str):
        class OptionImpl(MetricsOption):
            def fill(self, cfg: MetricsCfg) -> None:
                if not utils.is_empty_str(prefix):
                    cfg.prefix = prefix

        return OptionImpl()

    @staticmethod
    def with_metrics_http_schema(schema: str):
        class OptionImpl(MetricsOption):
            def fill(self, cfg: MetricsCfg) -> None:
                # only support "http" and "https"
                if schema in ["http", "https"]:
                    cfg.http_schema = schema

        return OptionImpl()

    # if not set, will report metrics.
    @staticmethod
    def enable_metrics():
        class OptionImpl(MetricsOption):
            def fill(self, cfg: MetricsCfg) -> None:
                cfg.enable_metrics = True

        return OptionImpl()

    # if not set, will not report metrics logs.
    @staticmethod
    def enable_metrics_log():
        class OptionImpl(MetricsOption):
            def fill(self, cfg: MetricsCfg) -> None:
                cfg.enable_metrics_log = True

        return OptionImpl()

    # set the interval of reporting metrics
    @staticmethod
    def with_report_interval_seconds(report_interval_seconds: float):
        class OptionImpl(MetricsOption):
            def fill(self, cfg: MetricsCfg) -> None:
                if report_interval_seconds > 1:
                    cfg.report_interval_seconds = report_interval_seconds

        return OptionImpl()

    @staticmethod
    def with_metrics_timeout_seconds(metrics_timeout_seconds: int):
        class OptionImpl(MetricsOption):
            def fill(self, cfg: MetricsCfg) -> None:
                cfg.http_timeout_seconds = metrics_timeout_seconds

        return OptionImpl()



import logging
import time
from typing import List, Optional, Dict
import requests
from requests import Response, Session

from byteplus_rec_core import constant, utils
from byteplus_rec_core.abtract_host_availabler import AbstractHostAvailabler, HostAvailabilityScore

log = logging.getLogger(__name__)

_DEFAULT_WINDOW_SIZE: int = 60
_DEFAULT_PING_SCHEMA = "http"
_DEFAULT_PING_URL_FORMAT: str = "{}://{}/predict/api/ping"
_DEFAULT_PING_TIMEOUT_SECONDS: float = 0.3
_DEFAULT_PING_INTERVAL_SECONDS: float = 1
_DEFAULT_FETCH_HOST_INTERVAL_SECONDS: float = 10


class Config(object):
    def __init__(self,
                 ping_url_format: str = _DEFAULT_PING_URL_FORMAT,
                 window_size: int = _DEFAULT_WINDOW_SIZE,
                 ping_timeout_seconds: float = _DEFAULT_PING_TIMEOUT_SECONDS,
                 ping_interval_seconds: float = _DEFAULT_PING_INTERVAL_SECONDS,
                 fetch_host_interval_seconds: float = _DEFAULT_FETCH_HOST_INTERVAL_SECONDS):
        self.ping_url_format = ping_url_format
        self.window_size = window_size
        if window_size < 0:
            self.window_size = _DEFAULT_WINDOW_SIZE
        self.ping_timeout_seconds = ping_timeout_seconds
        self.ping_interval_seconds = ping_interval_seconds
        self.fetch_host_interval_seconds = fetch_host_interval_seconds


class PingHostAvailabler(AbstractHostAvailabler):
    def __init__(self, default_hosts: Optional[List[str]] = None,
                 project_id: Optional[str] = None,
                 config: Optional[Config] = None):
        if config is None:
            config = Config()
        self._config: Config = config
        self._ping_http_cli: Session = requests.Session()
        self._host_window_map: Dict[str, _Window] = {}
        for host in default_hosts:
            self._host_window_map[host] = _Window(self._config.window_size)
        super().__init__(
            default_hosts,
            project_id,
            self._config.fetch_host_interval_seconds,
            self._config.ping_interval_seconds
        )
        return

    def do_score_hosts(self, hosts: List[str]) -> List[HostAvailabilityScore]:
        log.debug("[ByteplusSDK] do score hosts:'%s'", hosts)
        if len(hosts) == 1:
            return [HostAvailabilityScore(hosts[0], 0.0)]
        host_availability_scores = []
        for host in hosts:
            window = self._host_window_map[host]
            if window is None:
                window = _Window(self._config.window_size)
                self._host_window_map[host] = window
            success = utils.ping(self.project_id, self._ping_http_cli, self._config.ping_url_format,
                                 _DEFAULT_PING_SCHEMA, host, self._config.ping_timeout_seconds)
            window.put(success)
            host_availability_scores.append(HostAvailabilityScore(host, 1 - window.failure_rate()))
        return host_availability_scores


class _Window(object):
    def __init__(self, size: int):
        self.size: int = size
        self.head: int = size - 1
        self.tail: int = 0
        self.items: list = [True] * size
        self.failure_count: int = 0

    def put(self, success: bool) -> None:
        if not success:
            self.failure_count += 1
        self.head = (self.head + 1) % self.size
        self.items[self.head] = success
        self.tail = (self.tail + 1) % self.size
        removing_item = self.items[self.tail]
        if not removing_item:
            self.failure_count -= 1

    def failure_rate(self) -> float:
        return self.failure_count / self.size



#!/usr/bin/env python
# Learn more: https://github.com/kennethreitz/setup.py
import os

from codecs import open

from setuptools import setup, find_packages
import pkg_resources

here = os.path.abspath(os.path.dirname(__file__))

with open('requirements.txt') as f:
    install_requires = [str(req) for req in pkg_resources.parse_requirements(f)]

print(install_requires)

about = {}
with open(os.path.join(here, 'byteplus_rec_core', '__version__.py'), 'r', 'utf-8') as f:
    exec(f.read(), about)

with open('README.md', 'r', 'utf-8') as f:
    readme = f.read()

print(find_packages())

print(here)

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3",
    install_requires=install_requires,
    license=about['__license__'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    project_urls={
        'Documentation': 'https://docs.byteplus.com',
        'Source': 'https://github.com/byteplus-sdk/byteplus-sdk-python-rec-core',
    },
)



import datetime
from typing import Optional


class Options(object):
    def __init__(self):
        self.timeout: Optional[datetime.timedelta] = None
        self.request_id: Optional[str] = None
        self.headers: Optional[dict] = None
        self.queries: Optional[dict] = None
        self.server_timeout: Optional[datetime.timedelta] = None



import datetime
from abc import abstractmethod

from byteplus_rec_core.options import Options


class Option(object):
    @abstractmethod
    def fill(self, options: Options) -> None:
        raise NotImplementedError

    @staticmethod
    def conv_to_options(opts: tuple) -> Options:
        options: Options = Options()
        for opt in opts:
            opt.fill(options)
        return options

    # Specifies the timeout for this request
    @staticmethod
    def with_timeout(timeout: datetime.timedelta):
        class OptionImpl(Option):
            def fill(self, options: Options) -> None:
                options.timeout = timeout

        return OptionImpl()

    # Specify the request_id manually. By default,
    # the SDK generates a unique request_id for each request using the UUID
    @staticmethod
    def with_request_id(request_id: str):
        class OptionImpl(Option):
            def fill(self, options: Options) -> None:
                options.request_id = request_id

        return OptionImpl()

    # Add an HTTP header to the request.
    # In general, you do not need to care this.
    @staticmethod
    def with_http_header(key: str, value: str):
        class OptionImpl(Option):
            def fill(self, options: Options) -> None:
                if options.headers is None:
                    options.headers = {}
                options.headers[key] = value
        return OptionImpl()

    # Add an HTTP query to the request.
    # In general, you do not need to care this.
    @staticmethod
    def with_http_query(key: str, value: str):
        class OptionImpl(Option):
            def fill(self, options: Options) -> None:
                if options.queries is None:
                    options.queries = {}
                options.queries[key] = value

        return OptionImpl()

    # Specifies the maximum time it will take for
    # the server to process the request. The server will try to return
    # the result within this time, even if the task is not completed
    @staticmethod
    def with_server_timeout(timeout: datetime.timedelta):
        class OptionImpl(Option):
            def fill(self, options: Options) -> None:
                options.server_timeout = timeout

        return OptionImpl()



import copy
import json
import uuid
from abc import abstractmethod
import logging
import threading
import time
from typing import List, Optional, Dict
import requests
from requests import Response

from byteplus_rec_core import utils
from byteplus_rec_core.exception import BizException
from byteplus_rec_core import constant
from byteplus_rec_core.metrics.metrics import Metrics
from byteplus_rec_core.metrics.metrics_log import MetricsLog

log = logging.getLogger(__name__)

_FETCH_HOST_URL_FORMAT: str = "http://{}/data/api/sdk/host?project_id={}"
_HOST_AVAILABLE_SCORE_FORMAT: str = "host={}, score={}"
_DEFAULT_FETCH_HOST_INTERVAL_SECONDS: float = 10
_DEFAULT_SCORE_HOST_INTERVAL_SECONDS: float = 1


class HostAvailabilityScore:
    def __init__(self, host: Optional[str] = None, score: Optional[float] = None):
        self.host = host
        self.score = score

    def __str__(self):
        return _HOST_AVAILABLE_SCORE_FORMAT.format(self.host, self.score)


# class AvailablerConfig(object):
#     def __init__(self, default_hosts: Optional[List[str]] = None,
#                  project_id: Optional[str] = None,
#                  host_config: Optional[Dict[str, List[str]]] = None):
#         self.default_hosts = default_hosts
#         self.project_id = project_id
#         self.host_config = host_config


class AbstractHostAvailabler(object):
    def __init__(self, default_hosts: Optional[List[str]] = None,
                 project_id: Optional[str] = None,
                 fetch_host_interval_seconds: Optional[float] = _DEFAULT_FETCH_HOST_INTERVAL_SECONDS,
                 score_host_interval_seconds: Optional[float] = _DEFAULT_SCORE_HOST_INTERVAL_SECONDS):
        self.project_id = project_id
        self._default_hosts = default_hosts
        self._host_config = None
        self._abort: bool = False
        self._close_fetch_hosts_flag: bool = False
        self._fetch_hosts_thread = None
        self._fetch_host_interval_seconds = fetch_host_interval_seconds
        self._score_host_interval_seconds = score_host_interval_seconds
        self.init()

    def init(self):
        self.set_hosts(self._default_hosts)
        if not utils.is_empty_str(self.project_id):
            self._fetch_hosts_from_server()
            self._fetch_hosts_thread = threading.Thread(target=self._start_fetch_hosts_from_server)
            self._fetch_hosts_thread.start()
        threading.Thread(target=self._start_score_and_update_hosts).start()

    def set_hosts(self, hosts: List[str]):
        if hosts is None or len(hosts) == 0:
            raise BizException("host array is empty")
        self._default_hosts = hosts
        self._stop_fetch_hosts_from_server()
        self._score_and_update_hosts({"*": hosts})

    def _stop_fetch_hosts_from_server(self):
        # do not need to close fetch thread when setting default hosts
        # close fetch thread only when set_hosts was called by user
        if self._fetch_hosts_thread is not None:
            self._close_fetch_hosts_flag = True

    def _start_fetch_hosts_from_server(self):
        if self._close_fetch_hosts_flag or self._abort:
            return
        time.sleep(self._fetch_host_interval_seconds)
        self._fetch_hosts_from_server()
        self._start_fetch_hosts_from_server()
        return

    def _start_score_and_update_hosts(self):
        if self._abort:
            return
        time.sleep(self._score_host_interval_seconds)
        self._score_and_update_hosts(self._host_config)
        self._start_score_and_update_hosts()
        return

    def _fetch_hosts_from_server(self):
        url: str = _FETCH_HOST_URL_FORMAT.format(self._default_hosts[0], self.project_id)
        req_id: str = "fetch_" + str(uuid.uuid1())
        for i in range(3):
            rsp_host_config: Dict[str, List[str]] = self._do_fetch_hosts_from_server(req_id, url)
            if not rsp_host_config:
                continue
            if self._is_server_hosts_not_updated(rsp_host_config):
                MetricsLog.info(req_id,
                                "[ByteplusSDK][Fetch] hosts from server are not changed, project_id:{}, config: {}",
                                self.project_id, rsp_host_config)
                log.debug("[ByteplusSDK] hosts from server are not changed, config:'%s'", rsp_host_config)
                return
            if "*" not in rsp_host_config or rsp_host_config["*"] == []:
                metrics_tags = [
                    "type:no_default_hosts",
                    "project_id:" + self.project_id,
                    "url:" + utils.escape_metrics_tag_value(url),
                ]
                Metrics.counter(constant.METRICS_KEY_COMMON_WARN, 1, *metrics_tags)
                MetricsLog.warn(req_id,
                                "[ByteplusSDK][Fetch] no default value in hosts from server, project_id:{}, config: {}",
                                self.project_id, rsp_host_config)
                log.warning("[ByteplusSDK] no default value in hosts from server, config:'%s'", rsp_host_config)
                return
            self._score_and_update_hosts(rsp_host_config)
            return
        metrics_tags = [
            "type:fetch_host_fail_although_retried",
            "project_id:" + self.project_id,
            "url:" + utils.escape_metrics_tag_value(url),
        ]
        Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
        MetricsLog.warn(req_id,
                        "[ByteplusSDK][Fetch] fetch host from server fail although retried, project_id:{}, url: {}",
                        self.project_id, url)
        log.warning("[ByteplusSDK] fetch host from server fail although retried, url:'%s'", url)

    def _is_server_hosts_not_updated(self, new_host_config: Dict[str, List[str]]) -> bool:
        if self._host_config is None or new_host_config is None:
            return False
        for path in self._host_config:
            new_host: List[str] = new_host_config.get(path)
            old_host: List[str] = self._host_config.get(path)
            if old_host is None or new_host is None or len(old_host) == 0:
                return False
            if len(old_host) != len(new_host):
                return False
            if not set(old_host).issubset(set(new_host)):
                return False
        return True

    def _do_fetch_hosts_from_server(self, req_id: str, url: str) -> Dict[str, List[str]]:
        start = time.time()
        try:
            headers = {
                "Request-Id": req_id,
            }
            rsp: Response = requests.get(url, headers=headers, timeout=10)
            cost = int((time.time() - start) * 1000)
            if rsp.status_code == constant.HTTP_STATUS_NOT_FOUND:
                metrics_tags = [
                    "type:fetch_host_status_400",
                    "project_id:" + self.project_id,
                    "url:" + utils.escape_metrics_tag_value(url),
                ]
                Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
                log_format = "[ByteplusSDK][Fetch] fetch host from server return not found status project_id:{}, " \
                             "cost: {}ms"
                MetricsLog.warn(req_id, log_format, self.project_id, cost)
                log.warning("[ByteplusSDK] fetch host from server return not found status, cost:%dms", cost)
                return {}
            if rsp.status_code != constant.HTTP_STATUS_OK:
                metrics_tags = [
                    "type:fetch_host_not_ok",
                    "project_id:" + self.project_id,
                    "url:" + utils.escape_metrics_tag_value(url),
                ]
                Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
                log_format = "[ByteplusSDK][Fetch] fetch host from server return not ok status, project_id:{}, " \
                             "cost: {}ms, err:{}"
                MetricsLog.warn(req_id, log_format, self.project_id, cost, rsp.reason)
                log.warning("[ByteplusSDK] fetch host from server return not ok status, cost:%dms, err:'%s'",
                            cost, rsp.reason)
                return {}
            rsp_str: str = str(rsp.content)
            metrics_tags = [
                "project_id:" + self.project_id,
                "url:" + utils.escape_metrics_tag_value(url),
            ]
            Metrics.timer(constant.METRICS_KEY_REQUEST_TOTAL_COST, cost, *metrics_tags)
            Metrics.counter(constant.METRICS_KEY_REQUEST_COUNT, 1, *metrics_tags)
            log.debug("[ByteplusSDK] fetch host from server, cost:%dms, rsp:'%s'", cost, rsp_str)
            if rsp_str is not None and len(rsp_str) > 0:
                return json.loads(rsp.text)
            return {}
        except BaseException as e:
            cost = int((time.time() - start) * 1000)
            metrics_tags = [
                "type:fetch_host_fail",
                "project_id:" + self.project_id,
                "url:" + utils.escape_metrics_tag_value(url),
            ]
            Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
            log_format = "[ByteplusSDK][Fetch] fetch host from server err, project_id:{}, url:{}, cost:{}ms, err:{}"
            MetricsLog.warn(req_id, log_format, self.project_id, url, cost, e)
            log.warning("[ByteplusSDK] fetch host from server err, url:'%s', cost %dms, err:'%s'", url, cost, e)
            return {}

    def _score_and_update_hosts(self, host_config: Dict[str, List[str]]):
        log_id: str = "score_" + str(uuid.uuid1())
        hosts: List[str] = self._distinct_hosts(host_config)
        new_host_scores: List[HostAvailabilityScore] = self.do_score_hosts(hosts)
        MetricsLog.info(log_id, "[ByteplusSDK][Score] score hosts: project_id: {}, result:{}",
                        self.project_id, new_host_scores)
        log.debug("[ByteplusSDK] score hosts result: '%s'", new_host_scores)
        if new_host_scores is None or len(new_host_scores) == 0:
            metrics_tags = [
                "type:scoring_hosts_return_empty_list",
                "project_id:" + self.project_id,
            ]
            Metrics.counter(constant.METRICS_KEY_COMMON_ERROR, 1, *metrics_tags)
            MetricsLog.error(log_id, "[ByteplusSDK][Score] scoring hosts return an empty list, project_id:{}",
                             self.project_id)
            log.error("[ByteplusSDK] scoring hosts return an empty list")
            return
        new_host_config: Dict[str, List[str]] = self._copy_and_sort_host(host_config, new_host_scores)
        if self._is_host_config_not_update(self._host_config, new_host_config):
            MetricsLog.info(log_id, "[ByteplusSDK][Score] host order is not changed, project_id: {}, config:{}",
                            self.project_id, new_host_config)
            log.debug("[ByteplusSDK] host order is not changed, '%s'", new_host_config)
            return
        metrics_tags = [
            "type:set_new_host_config",
            "project_id:" + self.project_id,
        ]
        Metrics.counter(constant.METRICS_KEY_COMMON_INFO, 1, *metrics_tags)
        MetricsLog.info(log_id, "[ByteplusSDK][Score] set new host config:{}, old config: {}, project_id: {}",
                        new_host_config, self._host_config, self.project_id)
        log.warning("[ByteplusSDK] set new host config: '%s', old config: '%s'", new_host_config,
                    self._host_config)
        self._host_config = new_host_config

    @staticmethod
    def _distinct_hosts(host_config: Dict[str, List[str]]):
        host_set = set()
        for path in host_config:
            host_set.update(host_config[path])
        return list(host_set)

    @abstractmethod
    def do_score_hosts(self, hosts: List[str]) -> List[HostAvailabilityScore]:
        raise NotImplementedError

    @staticmethod
    def _copy_and_sort_host(host_config: Dict[str, List[str]], new_host_scores: List[HostAvailabilityScore]) -> \
            Dict[str, List[str]]:
        host_score_index = {}
        for host_score in new_host_scores:
            host_score_index[host_score.host] = host_score.score
        new_host_config = {}
        for path in host_config:
            new_hosts: List[str] = copy.deepcopy(host_config[path])
            # sort from big to small
            new_hosts = sorted(new_hosts, key=lambda s: host_score_index.get(s, 0.0), reverse=True)
            new_host_config[path] = new_hosts
        return new_host_config

    @staticmethod
    def _is_host_config_not_update(old_host_config: Dict[str, List[str]],
                                   new_host_config: Dict[str, List[str]]) -> bool:
        if old_host_config is None:
            return False
        if new_host_config is None:
            return True
        for path in old_host_config:
            new_hosts = new_host_config.get(path)
            old_hosts = old_host_config.get(path)
            if new_hosts != old_hosts:
                return False
        return True

    def get_hosts(self) -> List[str]:
        return self._distinct_hosts(self._host_config)

    def get_host(self, path: str) -> str:
        hosts = self._host_config.get(path)
        if hosts is None or len(hosts) == 0:
            return self._host_config.get("*")[0]
        return hosts[0]

    def shutdown(self):
        self._abort = True



from byteplus_rec_core import utils

from byteplus_rec_core.metrics import constant
from byteplus_rec_core.metrics.metrics_collector import MetricsCollector


class MetricsLog(object):
    @staticmethod
    def trace(log_id: str, log_format: str, *args):
        message: str = log_format.format(*args)
        MetricsCollector.emit_log(log_id, message, constant.LOG_LEVEL_TRACE, utils.current_time_millis())

    @staticmethod
    def debug(log_id: str, log_format: str, *args):
        message: str = log_format.format(*args)
        MetricsCollector.emit_log(log_id, message, constant.LOG_LEVEL_DEBUG, utils.current_time_millis())

    @staticmethod
    def info(log_id: str, log_format: str, *args):
        message: str = log_format.format(*args)
        MetricsCollector.emit_log(log_id, message, constant.LOG_LEVEL_INFO, utils.current_time_millis())

    @staticmethod
    def notice(log_id: str, log_format: str, *args):
        message: str = log_format.format(*args)
        MetricsCollector.emit_log(log_id, message, constant.LOG_LEVEL_NOTICE, utils.current_time_millis())

    @staticmethod
    def warn(log_id: str, log_format: str, *args):
        message: str = log_format.format(*args)
        MetricsCollector.emit_log(log_id, message, constant.LOG_LEVEL_WARN, utils.current_time_millis())

    @staticmethod
    def error(log_id: str, log_format: str, *args):
        message: str = log_format.format(*args)
        MetricsCollector.emit_log(log_id, message, constant.LOG_LEVEL_ERROR, utils.current_time_millis())

    @staticmethod
    def fatal(log_id: str, log_format: str, *args):
        message: str = log_format.format(*args)
        MetricsCollector.emit_log(log_id, message, constant.LOG_LEVEL_FATAL, utils.current_time_millis())



from abc import abstractmethod
from typing import List
from enum import Enum


class AbstractRegion(Enum):

    @abstractmethod
    def get_hosts(self) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def get_auth_region(self) -> str:
        raise NotImplementedError
